/*
	The MIT License
	
	Copyright (c) 2008-2009 Mark Walters
	
	Permission is hereby granted, free of charge, to any person obtaining a copy
	of this software and associated documentation files (the "Software"), to deal
	in the Software without restriction, including without limitation the rights
	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
	copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:
	
	The above copyright notice and this permission notice shall be included in
	all copies or substantial portions of the Software.
	
	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
	THE SOFTWARE.
*/

package com.yourpalmark.geoglobe
{
	import caurina.transitions.Tweener;
	
	import flash.display.Sprite;
	import flash.events.Event;
	import flash.events.MouseEvent;
	
	import mx.core.UIComponent;
	
	import org.papervision3d.core.math.Number3D;
	import org.papervision3d.events.InteractiveScene3DEvent;
	import org.papervision3d.objects.DisplayObject3D;
	import org.papervision3d.view.BasicView;
	
	public class GlobeView extends BasicView
	{
		protected var latitudeDegreeOffset:Number = 90;
		protected var longitudeDegreeOffset:Number = 0;
		
		protected var background:Sprite;
		protected var cameraTarget:DisplayObject3D;
		protected var globe:Globe;
		protected var globeMouseDiffX:Number = 0;
		protected var globeMouseDiffY:Number = 0;
		protected var mouseDown:Boolean = false;
		protected var geocoder:Geocoder;
		
		public function GlobeView( viewportWidth:Number=640, viewportHeight:Number=480, scaleToStage:Boolean=true, interactive:Boolean=true, cameraType:String="Free" )
		{
			//Setup the basic view.
			super( viewportWidth, viewportHeight, scaleToStage, interactive, cameraType );
			init();
			startRendering();
		}
		
		public function addMarkerByLocation( location:String ):void
		{
			geocoder.locate( location );
		}
		
		public function addMarkerByLatitudeLongitude( latitude:Number, longitude:Number, name:String=null ):void
		{
			trace( name );
			trace( latitude );
			trace( longitude );
			
			var marker:Marker = new Marker();
			marker.name = name;
			marker.addEventListener( InteractiveScene3DEvent.OBJECT_OVER, onMarkerOver );
			marker.addEventListener( InteractiveScene3DEvent.OBJECT_OUT, onMarkerOut );
			marker.addEventListener( InteractiveScene3DEvent.OBJECT_CLICK, onMarkerClick );
			placeMarker( marker, latitude, longitude );
			//changeView( marker );
		}
		
		public function clearMarkers():void
		{
			for each( var obj:DisplayObject3D in globe.children )
			{
				if( obj is Marker ) globe.removeChild( obj );
			}
		}
		
		protected function init():void
		{
			//Create background to fix rotation issues introduced when setting the view to interactive.
			background = new Sprite();
			addChildAt( background, 0 );
			
			//Create a null object for the camera to copy.
			cameraTarget = new DisplayObject3D();
			scene.addChild( cameraTarget );
			
			//Rotate the camera target to have the camera face America.
			cameraTarget.pitch( -30 );
			cameraTarget.yaw( 180 );
			
			//Create the globe.
			globe = new Globe();
			scene.addChild( globe );
			
			//Set the camera's focus and zoom. Defaults are 10 and 40 respectively.
			//60 on the zoom, along with 240 on the radius of the globe, seems to fix distortion issues.
			cameraAsCamera3D.focus = 10;
			cameraAsCamera3D.zoom = 60;
			
			//Listen to added to stage.
			addEventListener( Event.ADDED_TO_STAGE, onAddedToStage );
			
			//Listen for mouse events.
			addEventListener( MouseEvent.MOUSE_DOWN, onMouseDown );
			addEventListener( MouseEvent.MOUSE_UP, onMouseUp );
			addEventListener( MouseEvent.ROLL_OUT, onRollOut );
			
			//Initialize the geocoder.
			geocoder = new Geocoder();
			geocoder.addEventListener( GeocoderEvent.GEOCODER_SUCCESS, onGeocoderSuccess );
			geocoder.addEventListener( GeocoderEvent.GEOCODER_FAIL, onGeocoderFail );
			geocoder.addEventListener( GeocoderEvent.GEOCODER_UNAVAILABLE, onGeocoderUnavailable );
		}
		
		protected function placeMarker( marker:Marker, latitude:Number, longitude:Number ):void
		{
			//Translate the geo coordinates to 3D coordinates.
			var markerVector:Number3D = translateGeoCoords( latitude, longitude, globe.radius );
			
			marker.x = markerVector.x;
			marker.y = markerVector.y;
			marker.z = markerVector.z;
			
			//Align the marker with the globe's surface.
			marker.lookAt( DisplayObject3D.ZERO );
			
			globe.addChild( marker );
		}
		
		private function translateGeoCoords( latitude:Number, longitude:Number, radius:Number ):Number3D
		{
			//Convert latitude and longitude to radians.
			latitude = Math.PI * latitude / 180;
			longitude = Math.PI * longitude / 180;
			
			//Adjust latitude and longitude by radians.
			latitude -= ( latitudeDegreeOffset * ( Math.PI/180 ) ); // offset latitude by n degrees (in radians).
			longitude -= ( longitudeDegreeOffset * ( Math.PI/180 ) ); // offset longitude by n degrees (in radians).
			
			//Spherical to cartesian coordinate conversion formula.
			
			//----		r (radius)				----//
			//----		θ (theta) = latitude	----//
			//----		φ (phi) = longitude		----//
			
			//----		x = r sin θ cos φ		----//
			//----		y = r sin θ sin φ		----//
			//----		z = r cos θ				----//
			
			var x:Number = radius * Math.sin( latitude ) * Math.cos( longitude );
			var y:Number = radius * Math.sin( latitude ) * Math.sin( longitude );
			var z:Number = radius * Math.cos( latitude );
			
			//Switch z and y (since z is forward) (see the right-hand rule).
			return new Number3D( x, z, y );
		}
		
		override protected function onRenderTick( event:Event=null ):void
		{
			if( mouseDown )
			{
				//Set the target rotation properties for the camera based on the mouse position.
				var rotationX:Number = ( mouseY - globeMouseDiffY );
				var rotationY:Number = ( -mouseX - globeMouseDiffX );
				
				//Set a minimum and maximum amount that the camera can rotate around the x-axis.
				rotationX = Math.max( -190, Math.min( -125, rotationX ) );
				
				//Tween the null object.
				Tweener.addTween( cameraTarget, { rotationX:rotationX, rotationY:rotationY, time:3 } );
			}
			
			//Copy the null object's transform into the camera.
			cameraAsCamera3D.copyTransform( cameraTarget );
			
			//Reposition the camera away from the globe and null object.
			cameraAsCamera3D.moveBackward( 600 );
			cameraAsCamera3D.moveDown( 30 );
			
			//Render as usual
			super.onRenderTick( event );
		}
		
		protected function changeView( obj:DisplayObject3D ):void
		{
			Tweener.removeAllTweens();
			Tweener.addTween( cameraTarget, { rotationX:obj.rotationY, rotationY:obj.rotationX, time:3 } );
		}
		
		
		protected function onAddedToStage( event:Event ):void
		{
			background.graphics.clear();
			background.graphics.beginFill( 0x000000 );
			background.graphics.drawRect( 0, 0, stage.width, stage.height );
			background.graphics.endFill();
		}
		
		protected function onMouseDown( event:MouseEvent ):void
		{
			//Track the amount the mouse has moved.
			globeMouseDiffX = -(mouseX + cameraTarget.rotationY);
			globeMouseDiffY = mouseY - cameraTarget.rotationX;
			mouseDown = true;
		}
		
		protected function onMouseUp( event:MouseEvent ):void
		{
			mouseDown = false;
		}
		
		protected function onRollOut( event:MouseEvent ):void
		{
			mouseDown = false;
		}
		
		protected function onGeocoderSuccess( event:GeocoderEvent ):void
		{
			addMarkerByLatitudeLongitude( event.latitude, event.longitude, event.location );
		}
		
		protected function onGeocoderFail( event:GeocoderEvent ):void
		{
			trace( "GEOCODER FAIL" );
		}
		
		protected function onGeocoderUnavailable( event:GeocoderEvent ):void
		{
			trace( "GEOCODER UNAVAILABLE" );
		}
		
		protected function onMarkerOver( event:InteractiveScene3DEvent ):void
		{
			//Turn cache as bitmap on to fix mouse flicker issues on markers.
			viewport.containerSprite.cacheAsBitmap = true;
			
			//Show hand cursor.
			viewport.buttonMode = true;
		}
		
		protected function onMarkerOut( event:InteractiveScene3DEvent ):void
		{
			//Turn cache as bitmap back off.
			viewport.containerSprite.cacheAsBitmap = false;
			
			//Hide hand cursor.
			viewport.buttonMode = false;
		}
		
		protected function onMarkerClick( event:InteractiveScene3DEvent ):void
		{
			var marker:Marker = Marker( event.target );
			GeoGlobe( UIComponent( parent ).document ).selectedMarker.text = marker.name;
		}
		
	}
}